/*
 * This work is part of the White Rabbit project
 *
 * Copyright (C) 2012 CERN (www.cern.ch)
 * Author: Aurelio Colosimo <aurelio@aureliocolosimo.it>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 */
#include <stdint.h>
#include <errno.h>
#include <ppsi/ppsi.h>
#include "wrpc.h"
#include <common-fun.h>
#include "../proto-ext-whiterabbit/wr-api.h"
#include "../proto-ext-whiterabbit/wr-constants.h"

/* All of these live in wrpc-sw/include */
#include "minic.h"
#include "syscon.h"
#include "endpoint.h"
#include "softpll_ng.h"
#include "pps_gen.h"
#include "uart.h"
#include "rxts_calibrator.h"

extern int32_t cal_phase_transition[wr_num_ports];

int ptp_mode[2];
static int ptp_enabled[wr_num_ports];

static struct wr_operations wrpc_wr_operations = {
	.locking_enable = wrpc_spll_locking_enable,
	.locking_poll = wrpc_spll_locking_poll,
	.locking_disable = wrpc_spll_locking_disable,
	.enable_ptracker = wrpc_spll_enable_ptracker,

	.adjust_in_progress = wrpc_adjust_in_progress,
	.adjust_counters = wrpc_adjust_counters,
	.adjust_phase = wrpc_adjust_phase,

	.read_calib_data = wrpc_read_calibration_data,
	.calib_disable = wrpc_calibrating_disable,
	.calib_enable = wrpc_calibrating_enable,
	.calib_poll = wrpc_calibrating_poll,
	.calib_pattern_enable = wrpc_calibration_pattern_enable,
	.calib_pattern_disable = wrpc_calibration_pattern_disable,

	.enable_timing_output = wrpc_enable_timing_output,
};

/*ppi fields*/
static DSDefault  defaultDS;
static DSCurrent  currentDS;
static DSParent   parentDS;
static DSTimeProperties timePropertiesDS;
static struct pp_servo servo;
static struct wr_servo_state servo_state;

static struct wr_dsport wr_dsport[] = {
	{.ops = &wrpc_wr_operations,},
	{.ops = &wrpc_wr_operations,}
};
static DSPort     portDS[] = {
	{.ext_dsport = &wr_dsport[0],},
	{.ext_dsport = &wr_dsport[1],}
};

static int delay_ms[] = {PP_DEFAULT_NEXT_DELAY_MS,PP_DEFAULT_NEXT_DELAY_MS};
static int start_tics[wr_num_ports];

static struct pp_globals ppg_static; /* forward declaration */
static unsigned char __tx_buffer[wr_num_ports][PP_MAX_FRAME_LENGTH];
static unsigned char __rx_buffer[wr_num_ports][PP_MAX_FRAME_LENGTH];

/* despite the name, ppi_static is not static: tests/measure_t24p.c uses it */
struct pp_instance ppi_static[] = {
        // port 0
	{.glbs			= &ppg_static,
	.portDS			= &portDS[0],
	.n_ops			= &wrpc_net_ops,
	.t_ops			= &wrpc_time_ops,
	.vlans_array_len	= CONFIG_VLAN_ARRAY_SIZE,
	.proto			= PP_DEFAULT_PROTO,
	.mech			= PP_E2E_MECH, /* until changed by cfg */
	.iface_name		= "wr0",
	.port_name		= "wr0",
	.__tx_buffer		= __tx_buffer[0],
	.__rx_buffer		= __rx_buffer[0],},
        // port 1
	{.glbs			= &ppg_static,
	.portDS			= &portDS[1],
	.n_ops			= &wrpc_net_ops,
	.t_ops			= &wrpc_time_ops,
	.vlans_array_len	= CONFIG_VLAN_ARRAY_SIZE,
	.proto			= PP_DEFAULT_PROTO,
	.mech			= PP_E2E_MECH, /* until changed by cfg */
	.iface_name		= "wr1",
	.port_name		= "wr1",
	.__tx_buffer		= __tx_buffer[1],
	.__rx_buffer		= __rx_buffer[1],}
};

/* We now have a structure with all globals, and multiple ppi inside */
static struct pp_globals ppg_static = {
	.nlinks			= wr_num_ports,
	.pp_instances	= &ppi_static[0],
	.servo			= &servo,
	.defaultDS		= &defaultDS,
	.currentDS		= &currentDS,
	.parentDS		= &parentDS,
	.timePropertiesDS	= &timePropertiesDS,
	.global_ext_data	= &servo_state,
};

int wrc_ptp_init()
{
	int port;
	sdb_find_devices();
	uart_init_hw();

	pp_printf("PPSi for WRPC. Commit %s, built on " __DATE__ "\n",
		PPSI_VERSION);

	struct pp_globals *ppg = &ppg_static;

	for (port = 0; port < wr_num_ports; ++port) 
	{
		ptp_mode[port] = WRC_MODE_UNKNOWN;
		ppg->pp_instances[port] = ppi_static[port];
		ptp_enabled[port] = 0;
	}
	pp_init_globals(ppg, &__pp_default_rt_opts);

	return 0;
}

#define LOCK_TIMEOUT_FM (4 * TICS_PER_SECOND)
#define LOCK_TIMEOUT_GM (60 * TICS_PER_SECOND)

int wrc_ptp_set_mode(int mode, int port)
{
	uint32_t start_tics, lock_timeout = 0;
	struct pp_globals *ppg = &ppg_static;
	struct pp_instance *ppi[wr_num_ports];
	struct wr_dsport *wrp[wr_num_ports];

	for (port=0; port<wr_num_ports;port++) {
		ppi[port] = INST(ppg, port);
		wrp[port] = WR_DSPOR(ppi[port]);
	}

	typeof(ppg->rt_opts->clock_quality.clockClass) *class_ptr;
	int error = 0;
	/*
	 * We need to change the class in the default options.
	 * Unfortunately, ppg->rt_opts may be yet unassigned when this runs
	 */
	class_ptr = &__pp_default_rt_opts.clock_quality.clockClass;

	for (port=0; port<wr_num_ports;port++) {
		ptp_mode[port] = WRC_MODE_UNKNOWN;
		ptp_enabled[port] = 0;
		wr_servo_reset(ppi[port]);
		ppi[port]->cfg.ext = PPSI_EXT_WR; // Enable WR mode
	}

	switch (mode) {
	case WRC_MODE_GM:
	case WRC_MODE_ABSCAL: /* absolute calibration, gm-lookalike */
		for (port=0; port<wr_num_ports;port++) {
			wrp[port]->wrConfig = WR_M_ONLY;
			ppi[port]->role = PPSI_ROLE_MASTER;
			DSDEF(ppi[port])->clockQuality.clockClass = PP_CLASS_WR_GM_LOCKED;
			m1(ppi[port]); // port0 and por1 share one pp_global
		}
		*class_ptr = PP_CLASS_WR_GM_LOCKED;
		shw_pps_gen_unmask_output(1);
		spll_init(SPLL_MODE_GRAND_MASTER, 0, 1);
		lock_timeout = LOCK_TIMEOUT_GM;
		break;

	case WRC_MODE_MASTER:
		for (port=0; port<wr_num_ports;port++) {
			ppi[port]->role = PPSI_ROLE_MASTER;
			wrp[port]->wrConfig = WR_M_ONLY;
			DSDEF(ppi[port])->clockQuality.clockClass = PP_CLASS_DEFAULT;
			m1(ppi[port]);
		}
		*class_ptr = PP_CLASS_DEFAULT;
		shw_pps_gen_unmask_output(1);
		spll_init(SPLL_MODE_FREE_RUNNING_MASTER, 0, 1);
		lock_timeout = LOCK_TIMEOUT_FM;
		break;
	
	case WRC_MODE_CASCADED:
		for (port=1; port<wr_num_ports;port++) {
			ppi[port]->role = PPSI_ROLE_MASTER;
			wrp[port]->wrConfig = WR_M_AND_S;
			DSDEF(ppi[port])->clockQuality.clockClass = PP_CLASS_DEFAULT;
			m1(ppi[port]);
		}
		ppi[0]->role = PPSI_ROLE_SLAVE;
		wrp[0]->wrConfig = WR_M_AND_S;
		DSDEF(ppi[0])->clockQuality.clockClass = PP_CLASS_DEFAULT;

		*class_ptr = PP_CLASS_DEFAULT;
		shw_pps_gen_unmask_output(1);
		spll_init(SPLL_MODE_SLAVE, 0, 1);
		break;

	case WRC_MODE_SLAVE:
		for (port=0; port<wr_num_ports;port++) {
			ppi[port]->role = PPSI_ROLE_SLAVE;
			wrp[port]->wrConfig = WR_S_ONLY;
			DSDEF(ppi[port])->clockQuality.clockClass = PP_CLASS_DEFAULT;
			m1(ppi[port]);
		}
		*class_ptr = PP_CLASS_SLAVE_ONLY;
		spll_init(SPLL_MODE_SLAVE, 0, 1);
		break;
	}

	start_tics = timer_get_tics();

	pp_printf("Locking PLL");
	for (port=0; port<wr_num_ports;port++) {
		wrp[port]->ops->enable_timing_output(ppi[port], 0); /* later, wr_init chooses */
	}

	while (!spll_check_lock(0) && lock_timeout) {
		spll_update();
		timer_delay(TICS_PER_SECOND);
		if (timer_get_tics() - start_tics > lock_timeout) {
			pp_printf("\nLock timeout.");
			error = -ETIMEDOUT;
			break;
		}
		pp_printf(".");
	}
	pp_printf("\n");
	shw_pps_gen_enable_output(1);

	/* If we can't lock to the atomic/gps, we say it in the class */
	if (error && (mode == WRC_MODE_GM))
		*class_ptr = PP_CLASS_WR_GM_UNLOCKED;

	switch (mode) {		
	case WRC_MODE_GM:
	case WRC_MODE_ABSCAL: /* absolute calibration, gm-lookalike */
	case WRC_MODE_MASTER:
	case WRC_MODE_SLAVE:
		for (port=0; port<wr_num_ports;port++)
			ptp_mode[port] = mode;
		break;
	
	case WRC_MODE_CASCADED:
		ptp_mode[0] = WRC_MODE_SLAVE;
		for (port=1; port<wr_num_ports;port++)
			ptp_mode[port] = WRC_MODE_MASTER;
		break;
	}
	return error;
}

int wrc_ptp_get_mode(int port)
{
	return ptp_mode[port];
}

int wrc_ptp_sync_mech(int e2e_p2p_qry, int port)
{
	struct pp_instance *ppi = &ppi_static[port];
	int running;

	if (!CONFIG_HAS_P2P)
		return ppi->mech;

	switch(e2e_p2p_qry) {
	case PP_E2E_MECH:
	case PP_P2P_MECH:
		running = wrc_ptp_run(-1, port);
		wrc_ptp_run(0, port);
		ppi->mech = e2e_p2p_qry;
		wrc_ptp_run(running, port);
		return 0;
	default:
		return ppi->mech;
	};
}

int wrc_ptp_start(int port)
{
	struct pp_globals *ppg = &ppg_static;
	struct pp_instance *ppi;

	pp_printf("Port %d PTP start\n",port);
	// pp_init_globals(&ppg_static, &__pp_default_rt_opts);

	/* Call the state machine. Being it in "Initializing" state, make
	 * ppsi initialize what is necessary */
	ppi = INST(ppg, port);
	ppi->state = PPS_INITIALIZING;
	ppi->current_state_item = NULL;
	ppi->port_idx = port;
	ppi->frgn_rec_best = -1;

	delay_ms[port] = pp_state_machine(ppi, NULL, 0);
	start_tics[port] = timer_get_tics();
	WR_DSPOR(ppi)->linkUP = FALSE;
	wr_servo_reset(ppi);
	pp_timeout_setall(ppi);
	ptp_enabled[port] = 1; 
	return 0;
}

int wrc_ptp_stop(int port)
{
	struct pp_globals *ppg = &ppg_static;
	struct pp_instance *ppi;
	struct wr_dsport *wrp;

	ppi = INST(ppg, port);
	wrp = WR_DSPOR(ppi);
	pp_printf("Port %d PTP stop\n",port);
	/* Moving fiber: forget about this parent (FIXME: shouldn't be here) */
	if(port==0) { 
		wrp->parentWrConfig = wrp->parentWrModeOn = 0;
		memset(ppi->frgn_master, 0, sizeof(ppi->frgn_master));
		ppi->frgn_rec_num = 0;		  /* no known master */
	}
	ptp_enabled[port] = 0;
	wr_servo_reset(ppi);
	// pp_close_globals(&ppg_static);  -- this function does nothing

	return 0;
}

int wrc_ptp_run(int start_stop_query, int port)
{
	switch(start_stop_query) {
	case 0:
		wrc_ptp_stop(port);
		return 0;
	case 1:
		wrc_ptp_start(port);
		return 0;
	default:
		return ptp_enabled[port];
	}
}

/* this returns whether or not the function did any work */
int wrc_ptp_update()
{
	int rv=0;
	int recv[wr_num_ports];
	int port;
	struct pp_globals *ppg = &ppg_static;
	struct pp_instance *ppi[wr_num_ports];

	for(port = 0; port<wr_num_ports; port++) 
	{
		if (!ptp_enabled[port])
			continue;
		ppi[port] = INST(ppg, port);
		recv[port] = __recv_and_count(ppi[port], ppi[port]->rx_frame, PP_MAX_FRAME_LENGTH - 4,
				 &(ppi[port]->last_rcv_time));

		if ((!recv[port]) && (timer_get_tics() - start_tics[port] < delay_ms[port]))
			continue;

		if (!recv[port]) {
			/* Nothing received, but timeout elapsed */
			start_tics[port] = timer_get_tics();
			delay_ms[port] = pp_state_machine(ppi[port], NULL, 0);
		} else {
			delay_ms[port] = pp_state_machine(ppi[port], ppi[port]->rx_ptp, recv[port]);
		}
		rv = 1;
	}

	return rv;
}
